# The core BirDePy functions estimate, probability and forecast are contained
# in 'interface' modules which must be imported.
from birdepy.interface_estimate import estimate
from birdepy.interface_probability import probability
from birdepy.interface_forecast import forecast
# The core BirDePy functions simulate.discrete and simulate.continuous are
# imported together
from birdepy import simulate
# Note that the gpu_functions module is not imported by default



from birdepy.interface_dnm import bld_ll_fun as bld_ll_fun
import functools
import numpy as np
import birdepy.utility as ut


def line_search_ll(direction, data, p0, likelihood, known_p,
                   idx_known_p, model, z_trunc, p_bounds, con, opt_method,
                   options):
    """
    Returns a scalar that determines the maximum of 'likelihood' along
    the subspace defined by 'direction' and point 'p0' satisfying constraints
    defined by 'p_bounds' and 'con'.
    """
    def upper_bound(a, index):
        return p_bounds[index][1] - p0[index] - a * direction[index]

    def lower_bound(a, index):
        return p0[index] + a * direction[index] - p_bounds[index][0]

    new_con = []
    if type(con) == dict:
        # If there is one constraint specified
        new_con.append({'type': con['type'],
                        'fun': lambda a:
                        con['fun'](p0 + np.multiply(a, direction))})
    else:
        # If there is strictly more than one or 0 constraints specified
        for c in con:
            new_con.append({'type': c['type'],
                            'fun': lambda a:
                            c['fun'](p0 + np.multiply(a, direction))})

    for idx in range(len(p_bounds)):
        new_con.append(
            {'type': 'ineq', 'fun': functools.partial(upper_bound, index=idx)})
        new_con.append(
            {'type': 'ineq', 'fun': functools.partial(lower_bound, index=idx)})

    new_con = tuple(new_con)

    ll = bld_ll_fun(data, likelihood, model, z_trunc, options)

    def error_fun(a):
        p_prop = p0 + np.multiply(a, direction)
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return -ll(param)

    a_opt = ut.minimize_(error_fun, 1e-6, None, new_con, opt_method, options).x

    return a_opt


def help_bld_aug(udh, sorted_data, j_tol, h_tol, z_trunc):
    """
    Creates set of augmented data using a sorted data set. Is set up so that
    expected values of up jumps, down jumps and holding times corresponding
    to each transition z0 to zt only needs to be computed once.
    """
    z_min, z_max = z_trunc
    aug_data = {}
    for t in sorted_data:
        for zz in sorted_data[t]:
            z0 = zz[0]
            zt = zz[1]
            lower_z = min(z0, zt)
            higher_z = max(z0, zt)
            for z in range(max(lower_z, z_min + 1), higher_z):
                udh_ = udh(z0, zt, t, z)
                if z in aug_data:
                    aug_data[z][:] += sorted_data[t][zz]*udh_
                else:
                    aug_data[z] = sorted_data[t][zz]*udh_
            z = higher_z
            while z < z_max:
                udh_ = udh(z0, zt, t, z)
                if z in aug_data:
                    aug_data[z][:] += sorted_data[t][zz]*udh_
                else:
                    aug_data[z] = sorted_data[t][zz]*udh_
                if udh_[0] < j_tol and udh_[1] < j_tol and udh_[2] < h_tol:
                    break
                else:
                    z += 1
            z = lower_z - 1
            while z > z_min:
                udh_ = udh(z0, zt, t, z)
                if z in aug_data:
                    aug_data[z][:] += sorted_data[t][zz]*udh_
                else:
                    aug_data[z] = sorted_data[t][zz]*udh_
                if udh_[0] < j_tol and udh_[1] < j_tol and udh_[2] < h_tol:
                    break
                else:
                    z -= 1
    return aug_data


def help_bld_aug_2(udh, z0, zt, aug_data, j_tol, h_tol, z_trunc):
    """
    Updates augmented data with new values corresponding to the
    transition from z0 to zt according to the udh function
    (which generates expected values).
    """
    z_min, z_max = z_trunc
    lower_z = min(z0, zt)
    higher_z = max(z0, zt)
    for z in range(max(lower_z, z_min + 1), higher_z):
        udh_ = udh(z)
        if z in aug_data:
            aug_data[z][:] += udh_
        else:
            aug_data[z] = udh_
    z = higher_z
    while z < z_max:
        udh_ = udh(z)
        if z in aug_data:
            aug_data[z][:] += udh_
        else:
            aug_data[z] = udh_
        if udh_[0] < j_tol and udh_[1] < j_tol and udh_[2] < h_tol:
            break
        else:
            z += 1
    z = lower_z - 1
    while z > z_min:
        udh_ = udh(z)
        if z in aug_data:
            aug_data[z][:] += udh_
        else:
            aug_data[z] = udh_
        if udh_[0] < j_tol and udh_[1] < j_tol and udh_[2] < h_tol:
            break
        else:
            z -= 1
    return aug_data




import numpy as np
import birdepy.utility as ut
from birdepy.interface_probability import probability
import warnings


def bld_ll_fun(data, likelihood, model, z_trunc, options):
    """
    Builds a log-likelihood function using the alternative methods for approximating
    transition probabilities.

    This function is used in :func:`bd.interface_dnm.discrete_est_dnm()` (where
    it is maximized).
    """
    # A different data structure is used depending on the likelihood method.
    # This is because for some likelihood methods each call of the function 'probability' naturally
    # generate sets of transition probabilities, while for others only a single transition
    # probability is generated.
    if likelihood in ['da', 'gwa', 'gwasa', 'ilt', 'oua']:
        def ll_fun(param):
            ll = 0
            for i in data:
                pr = probability(i[0], i[1], i[2], param, method=likelihood,
                                 model=model, z_trunc=z_trunc,
                                 options=options)[0][0]
                if np.isnan(pr):
                    warnings.warn("Computation of a transition probability has failed and been "
                                  "replaced by 1e-100. The results may be unreliable. ",
                                  category=RuntimeWarning)
                    ll += data[i] * np.log(1e-100)
                else:
                    ll += data[i] * np.log(np.maximum(1e-100, pr))
            return ll
    elif likelihood in ['Erlang', 'expm', 'uniform']:
        sorted_data = ut.data_sort_2(data)
        z_min, z_max = z_trunc

        def ll_fun(param):
            ll = 0
            for t in sorted_data:
                p_mat = probability(np.arange(z_min, z_max + 1, 1),
                                    np.arange(z_min, z_max+1, 1), t, param,
                                    method=likelihood, model=model,
                                    z_trunc=z_trunc, options=options)
                for i in sorted_data[t]:
                    ll += sorted_data[t][i] * np.log(
                        np.maximum(1e-100, p_mat[i[0] - z_min, i[1] - z_min]))
            return ll

    else:
        raise TypeError("Argument likelihood has an unknown value.")

    return ll_fun


def discrete_est_dnm(data, likelihood, model, z_trunc, idx_known_p, known_p,
                     p0, p_bounds, con, opt_method, options):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using direct numerical maximization of
    approximate likelihood functions.
    See :ref:`here <Direct Numerical Maximization>` for more information.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'dnm'::

        bd.estimate(t_data, p_data, p0, p_bounds, framework='dnm', likelihood='expm', z_trunc=())

    The parameters associated with this framework (listed below) can be
    accessed using kwargs in :func:`birdepy.estimate()`. See documentation of
    :func:`birdepy.estimate` for the main arguments.

    Parameters
    ----------
    likelihood : string, optional
        Likelihood approximation method. Should be one of
        (alphabetical ordering):

            - 'da' (see :ref:`here <birdepy.probability(method='da')>`)
            - 'Erlang' (see :ref:`here <birdepy.probability(method='Erlang')>`)
            - 'expm'  (default) (see :ref:`here <birdepy.probability(method='expm')>`)
            - 'gwa' (see :ref:`here <birdepy.probability(method='gwa')>`)
            - 'gwasa' (see :ref:`here <birdepy.probability(method='gwasa')>`)
            - 'ilt' (see :ref:`here <birdepy.probability(method='ilt')>`)
            - 'oua' (see :ref:`here <birdepy.probability(method='oua')>`)
            - 'uniform' (see :ref:`here <birdepy.probability(method='uniform')>`)
        The links point to the documentation of the relevant `method` in
        :func:`birdepy.probability`. The arguments associated with each of
        these methods may be used as a kwarg in :func:`birdepy.estimate()`
        when `likelihood` is set to use the method.

    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default\
        ``z_trunc=[z_min, z_max]`` where
        ``z_min=max(0, min(p_data) - 2*z_range)``
        and ``z_max=max(p_data) + 2*z_range`` where
        ``z_range=max(p_data)-min(p_data)``. Only applicable to `likelihood`
        methods 'Erlang', 'expm' and 'uniform'.

    Examples
    --------
    Simulate a sample path and estimate the parameters using the various
    likelihood approximation methods: ::

        import birdepy as bd
        t_data = list(range(100))
        p_data = bd.simulate.discrete([0.75, 0.25, 0.02, 1], 'Ricker', 10, t_data,
                                      survival=True, seed=2021)
        for likelihood in ['da', 'Erlang', 'expm', 'gwa', 'gwasa', 'ilt', 'oua', 'uniform']:
            likelihood = 'gwasa'
            est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[1e-6,1], [1e-6,1], [1e-6, 0.1]],
                             framework='dnm', model='Ricker', idx_known_p=[3], known_p=[1],
                              likelihood=likelihood)
            print('dnm estimate using', likelihood, 'is', est.p, ', with standard errors',
                  est.se, 'computed in ', est.compute_time, 'seconds.')

    Outputs: ::

        dnm estimate using da is [0.742967758293063, 0.21582082476775125, 0.022598554340938885] , with standard errors [0.16633436 0.03430501 0.00436483] computed in  12.29400086402893 seconds.
        dnm estimate using Erlang is [0.7478158070214841, 0.2150341741826537, 0.022748822721356705] , with standard errors [0.16991822 0.0345762  0.00435054] computed in  0.9690003395080566 seconds.
        dnm estimate using expm is [0.7477255598292176, 0.2150476994316206, 0.022745305129350565] , with standard errors [0.16904095 0.03443197 0.00433563] computed in  1.6919987201690674 seconds.
        dnm estimate using gwa is [0.6600230500097711, 0.16728663936008945, 0.02512248420514078] , with standard errors [0.14248815 0.02447161 0.00488879] computed in  37.52255415916443 seconds.
        dnm estimate using gwasa is [0.6604981297820195, 0.16924607541398484, 0.02492054535741541] , with standard errors [0.14244908 0.02485465 0.00488222] computed in  0.8699958324432373 seconds.
        dnm estimate using ilt is [0.7466254648849691, 0.21415145383850764, 0.022794996238547492] , with standard errors [0.10187377 0.03137803        nan] computed in  1185.0924031734467 seconds.
        dnm estimate using oua is [0.5000083585920406, 0.5, 0.05] , with standard errors [       nan        nan 0.01961143] computed in  3.466001272201538 seconds.
        dnm estimate using uniform is [0.7477293759434092, 0.215047068344254, 0.022745437226772615] , with standard errors [0.16900378 0.03443071 0.00433504] computed in  3.275972366333008 seconds.

    The constraint ``con={'type': 'ineq', 'fun': lambda p: p[0]-p[1]}`` ensures
    that p[0] > p[1] (i.e., rate of spread greater than recovery rate).

    How well methods perform varies from case to case. In this instance most methods perform well,
    while some throw errors but return useful output regardless, and some fail altogether.

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """
    # Build the log-likelihood function assuming all parameters are known
    pre_ll_fun = bld_ll_fun(data, likelihood, model, z_trunc, options)

    # Adjust the log-likelihood function to account for unknown parameters and
    # negate it so that the minimize function performs maximisation
    def error_fun(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return -pre_ll_fun(param)

    # The actual log-likelihood is needed to determine the covariance
    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll_fun(param)

    # Obtain parameter estimates
    opt = ut.minimize_(error_fun, p0, p_bounds, con, opt_method, options)

    # Obtain covariance estimates
    if p0.size > 1:
        try:
            cov = np.linalg.inv(-ut.Hessian(ll, opt.x, p_bounds))
        except:
            cov = 'Covariance matrix could not be determined.'
    else:
        try:
            cov = -1/ut.Hessian(ll, opt.x, p_bounds)
        except:
            cov = 'Covariance matrix could not be determined.'

    return opt, cov


def continuous_est_dnm(t_data, p_data, p0, b_rate, d_rate, p_bounds, con,
                       known_p, idx_known_p, opt_method, options):
    """Parameter estimation for continuously  observed continuous-time
    birth-and-death processes using direct numerical maximization of
    the likelihood.
    See :ref:`here <Continuously Observed Data>` for more information.

    To use this function call :func:`birdepy.estimate` with `scheme` set to
    'continuous'::

        bd.estimate(t_data, p_data, p0, p_bounds, scheme='continuous')


    Examples
    --------
    Simulate a continuous sample path and estimate the parameters.

    Import BirDePy: ::

        import birdepy as bd

    Simulate some synthetic data: ::

        t_data, p_data = bd.simulate.continuous([0.75, 0.25, 0.02, 1], 'Ricker', 10,
                                                100, survival=True, seed=2021)

    Estimate: ::

        est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[0,1], [0,1], [0, 0.1]],
                          model='Ricker', idx_known_p=[3], known_p=[1], scheme='continuous')
        print(est.p)

    Outputs: ::

        [0.7603171062895576, 0.2514810854871476, 0.020294342655751033]

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    --------


    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Wolff, R. W. (1965). Problems of statistical inference for birth and
     death queuing models. Operations Research, 13(3), 343-357.
    """
    # Convert data into the format of a list of lists
    if type(t_data[0]) != list:
        t_data = [t_data]
    if type(p_data[0]) != list:
        p_data = [p_data]

    # Create a list containing inter-event (or "holding") times
    holding_times = []
    for idx in range(len(t_data)):
        holding_times.append(np.array(t_data[idx][1:]) -
                             np.array(t_data[idx][0:-1]))

    # Create a list containing population sizes (or states) corresponding to holding_times
    flattened_p_data = [item for sublist in p_data for item in sublist]

    # Create an array of all observed population sizes
    states_visited = np.arange(np.amin(flattened_p_data),
                               np.amax(flattened_p_data) + 1, 1)

    # Create arrays to store holding times and counts of up and down jumps for each population size
    holding_t_data = np.zeros(len(states_visited))
    downward_jump_counts = np.zeros(len(states_visited))
    upward_jump_counts = np.zeros(len(states_visited))

    # Determine holding times and counts of up and down jumps for each population size
    for idx1 in range(len(t_data)):
        state_record_temp = np.array(p_data[idx1])
        holding_times_temp = np.array(holding_times[idx1])
        for idx2, state in enumerate(states_visited):
            holding_t_data[idx2] += \
                np.sum(holding_times_temp[state_record_temp[:-1] == state])
            downward_jump_counts[idx2] += \
                np.sum(np.less(state_record_temp[1:], state_record_temp[0:-1])
                       & (state_record_temp[0:-1] == state))
            upward_jump_counts[idx2] += \
                np.sum(np.greater(state_record_temp[1:], state_record_temp[0:-1])
                       & (state_record_temp[0:-1] == state))

    # Determine the log-likelihood function (e.g., Equation (2) in reference [3])
    def pre_ll(param):
        ll_ = 0
        for idx_, state_ in enumerate(states_visited):
            if b_rate(state_, param) > 0:
                ll_ += np.log(b_rate(state_, param)) * upward_jump_counts[idx_]
            if d_rate(state_, param) > 0:
                ll_ += np.log(d_rate(state_, param)) * downward_jump_counts[idx_]
            ll_ -= (b_rate(state_, param) + d_rate(state_, param)) * holding_t_data[idx_]
        return ll_

    # Adjust the log-likelihood function to account for unknown parameters and
    # negate it so that the minimize function performs maximisation
    def error_fun(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return -pre_ll(param)

    # The actual log-likelihood is needed to determine the covariance
    def ll(p_prop):
        param = ut.p_bld(p_prop, idx_known_p, known_p)
        return pre_ll(param)

    # Obtain parameter estimates
    opt = ut.minimize_(error_fun, p0, p_bounds, con, opt_method, options)

    # Obtain covariance estimates
    if p0.size > 1:
        try:
            cov = np.linalg.inv(-ut.Hessian(ll, opt.x, p_bounds))
        except:
            cov = 'Covariance matrix could not be determined.'
    else:
        try:
            cov = -1/ut.Hessian(ll, opt.x, p_bounds)
        except:
            cov = 'Covariance matrix could not be determined.'

    return opt, cov



from __future__ import print_function, absolute_import
import numpy as np
from numba import cuda
import math
from numba.cuda.random import create_xoroshiro128p_states
from numba.cuda.random import xoroshiro128p_uniform_float64 as rand
from collections import Counter
import sys


def discrete(param, model, z0, t, k=1, survival=False, seed=None):
    """Simulation of continuous-time birth-and-death processes at time 't'
    using CUDA.

    Parameters
    ----------
    param : array_like
        The parameters governing the evolution of the birth-and-death
        process to be simulated.
        Array of real elements of size (m,), where ‘m’ is the number of
        parameters.
        These must be in the order given :ref:`here
        <Birth-and-death Processes>`).

    model : string
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst' (default)
            - 'Ricker'
            - 'Hassell'
            - 'MS-S'
            - 'Moran'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'M/M/1'
            - 'M/M/inf'
            - 'loss-system'
         Custom models are not available for this function. See :func:`birdepy.simulate.discrete()`
         for custom models.

    z0: int or callable
        The initial population size for each sample path.

    t : scalar
        The time at which the simulated birth-and-death is observed.

    k : int, optional
        The number of sample paths to be simulated.

    survival : bool, optional
        If set to True, then the simulated sample paths are conditioned to
        have a positive population size at the final observation time.  Since This
        can greatly increase computation time.

    seed : int, optional
        Seed for simulation.

    Return
    -------
    out : array_like
        A list containing sampled population size observations at
        time `t`, generated according to `model`.

    Examples
    --------
    Simulating 10 ** 8 sample paths of an M/M/inf queue with
    service rates 0.4 and arrival rate 0.2, with 10 items initially in the queue,
    observed at time 1.0: ::

        from birdepy import gpu_functions as bdg
        bdg.discrete([0.2, 0.4], 'M/M/inf', 10, 1.0, k=10**8)
                     array([8, 6, 3, ..., 8, 7, 9], dtype=int64)

    Notes
    -----
    This function requires a compatible Nvidia graphics processing unit and
    drivers to be installed.

    The packages `Numba` and `cudatoolkit` also need to be installed.

    If you use this function for published work, then please cite [1].

    Sample paths are generated using a discrete-event simulation algorithm.
    See, for example, Algorithm 5.8 in [2].

    For a text book treatment on the theory of birth-and-death processes
    see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067. 

    .. [2] Kroese, D.P., Taimre, T. and Botev, Z.I. (2013) Handbook of Monte
     Carlo methods. John Wiley & Sons.

    .. [3] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.
    """
    # Convert parameters to tuple as this is accepted by the numba/cuda functions
    param = tuple(param)
    threads_per_block = 1024
    # Determine the number of blocks such that the number of samples is at least k
    blocks = int(1 + k / threads_per_block)
    # Determine a seed for the random number generator
    if seed is None:
        rng = np.random.default_rng()
        seed = rng.integers(sys.maxsize)
    # Initialise the random number generator
    rng_states = create_xoroshiro128p_states(threads_per_block * blocks, seed=seed)
    out = np.zeros(threads_per_block * blocks, dtype=np.int64)
    # A different function is called depending on which model is of interest. The numba/cuda
    # functions do not accept functions as input (at time of writing) so we cannot pass birth and
    # death rate functions as we do for the CPU version of this code bd.simulate.discrete().
    if model == 'Verhulst':
        discrete_verhulst[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Ricker':
        discrete_ricker[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Hassell':
        discrete_hassell[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'MS-S':
        discrete_mss[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Moran':
        discrete_moran[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'pure-birth':
        discrete_pb[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'pure-death':
        discrete_pd[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'Poisson':
        discrete_poisson[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'linear':
        discrete_linear[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'linear-migration':
        discrete_lm[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'M/M/1':
        discrete_mm1[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'M/M/inf':
        discrete_mminf[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    elif model == 'loss-system':
        discrete_loss[blocks, threads_per_block](out, rng_states, param, z0, t, survival)
        return out
    else:
        raise TypeError("Argument 'model' has an unknown value.")


def probability(z0, zt, t, param, model, k=10**6, seed=None):
    """Transition probabilities for continuous-time birth-and-death processes
    generated using Monte Carlo on a GPU.

    Parameters
    ----------
    z0 : array_like
        States of birth-and-death process at time 0.

    zt : array_like
        States of birth-and-death process at time(s) `t`.

    t : array_like
        Elapsed time(s) (if has size greater than 1, then must be increasing).

    param : array_like
        The parameters governing the evolution of the birth-and-death
        process.
        Array of real elements of size (n,), where ‘n’ is the number of
        parameters.
        These must be in the order given :ref:`here
        <Birth-and-death Processes>`).

    model : string, optional
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst' (default)
            - 'Ricker'
            - 'Hassell'
            - 'MS-S'
            - 'Moran'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'MS-S'
            - 'M/M/inf'
            - 'loss-system'
         Custom models are not available for this function. See :func:`birdepy.probability()`
         for custom models.

    k: int, optional
        Minimum number of samples used to generate each probability estimate.
        (Actual number of samples will usually be higher due to the way memory
        is allocated on GPU.)
        The total number of samples used will be at least z0.size * k.

    seed : int, optional
        Seed for simulations.

    Return
    -------
    transition_probability : ndarray
        An array of transition probabilities. If t has size bigger than 1,
        then the first coordinate corresponds to `t`, the second coordinate
        corresponds to `z0` and the third coordinate corresponds to `zt`;
        for example if `z0=[1,3,5,10]`, `zt=[5,8]` and `t=[1,2,3]`, then
        `transition_probability[2,0,1]` corresponds to
        P(Z(3)=8 | Z(0)=1).
        If `t` has size 1 the first coordinate corresponds to `z0` and the second
        coordinate corresponds to `zt`.

    Examples
    --------
    Estimate transition probabilities for a Moran model: ::

        from birdepy import gpu_functions as bdg
        param = (210, 20, 0.002, 0, 100)
        t = 0.2
        z0 = [50, 60]
        zt = [55, 56, 57, 58, 59,60]
        bdg.probability(z0, zt, t, param, 'Moran', 10**6)


    Outputs: ::

        array([[3.09160e-02, 5.43120e-02, 8.09760e-02, 1.05968e-01, 1.23203e-01,1.27453e-01],
        [0.00000e+00, 0.00000e+00, 0.00000e+00, 0.00000e+00, 0.00000e+00, 6.30000e-05]])

    Notes
    -----
    This function requires a compatible Nvidia graphics processing unit and
    drivers to be installed.

    The packages `Numba` and `cudatoolkit` also need to be installed.

    If you use this function for published work, then please cite [1].

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`


    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """

    # Convert z0, zt, and t to common format used below
    if np.isscalar(z0):
        z0 = np.array([z0])
    else:
        z0 = np.array(z0)
    if np.isscalar(zt):
        zt = np.array([zt])
    else:
        zt = np.array(zt)
    if np.isscalar(t):
        t = np.array([t])
    else:
        t = np.array(t)

    # Since bdg.discrete() can only accept a single time point we need to add an extra loop if more
    # than one time point is requested. This is easiest to achieve by diverting into different code
    # blocks depending on the number of time points.
    if t.size == 1:
        # Initialize an array to store output
        output = np.zeros((z0.size, zt.size))
        # Perform simulations using requested initial conditions
        for idx1, _z0 in enumerate(z0):
            sim = discrete(param, model, _z0, t[0], k, False, seed)
            # For each simulation from a specific initial condition count the number of
            # trajectories which fall into each terminal state
            counts = Counter(sim)
            # Fill in the output array according to requested terminal states
            for idx2, _zt in enumerate(zt):
                output[idx1, idx2] = counts[_zt]/k
    else:
        # Initialize an array to store output
        output = np.zeros((t.size, z0.size, zt.size))
        # Loop over requested times (note all simulations start at time 0)
        for idx3, _t in enumerate(t):
            # Perform simulations using requested initial conditions
            for idx1, _z0 in enumerate(z0):
                sim = discrete(param, model, _z0, _t, k, False, seed)
                # For each simulation from a specific intitial condition count the number of
                # trajectories which fall into each terminal state
                counts = Counter(sim)
                # Fill in the output array according to requested terminal states
                for idx2, _zt in enumerate(zt):
                    output[idx3, idx1, idx2] = counts[_zt] / k
    return output


@cuda.jit
def discrete_verhulst(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process Verhulst model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * (1 - p[2]*z) * z

    def d_rate(z):
        return p[1] * (1 + p[3]*z) * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_ricker(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process Ricker model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * math.exp(-(p[2] * z) ** p[3]) * z

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_bh(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process Beverton-Holt model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * p[2] / (z + p[2])

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_hassell(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process Hassell model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z / (1 + z * p[2]) ** p[3]

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_mss(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process MaynardSmith-Slatkin model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z / (1 + (z * p[2]) ** p[3])

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_moran(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process Moran model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return (p[4] - z) * (p[0] * z * (1 - p[2]) / p[4] +
                                     p[1] * (p[4] - z) * p[3] / p[4]) \
                       / p[4]

    def d_rate(z):
        return z * (p[2] * (p[4] - z) * (1 - p[3]) / p[4] +
                            p[0] * z * p[2] / p[4]) \
                       / p[4]

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_pb(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time pure birth process model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z

    def d_rate(z):
        return 0

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_pd(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time pure death process model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return 0

    def d_rate(z):
        return p[0] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_poisson(out, rng_states, p, z0, time, survival):
    """
    Simulation of Poisson process model at discrete observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0]

    def d_rate(z):
        return 0

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop



@cuda.jit
def discrete_linear(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process linear model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_lm(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process linear-migration model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * z + p[2]

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_mm1(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process single-server queue model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0]

    def d_rate(z):
        return p[1] * (z > 0)

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_mminf(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process infinite-server queue model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0]

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop


@cuda.jit
def discrete_loss(out, rng_states, p, z0, time, survival):
    """
    Simulation of continuous-time birth-and-death process loss-system model at discrete
    observation times.

    This function is used by :func:`bd.gpu_functions.discrete()`.
    """
    thread_id = cuda.grid(1)

    def b_rate(z):
        return p[0] * (z < p[2])

    def d_rate(z):
        return p[1] * z

    while True:
        pop = z0
        rate = b_rate(pop) + d_rate(pop)
        if rate == 0:
            next_event_time = math.inf
        else:
            next_event_time = -math.log(rand(rng_states, thread_id)) / rate
        while next_event_time <= time:
            if (rand(rng_states, thread_id) * rate) <= b_rate(pop):
                pop += 1
            else:
                pop -= 1
            rate = b_rate(pop) + d_rate(pop)
            if rate > 0:
                next_event_time += -math.log(rand(rng_states, thread_id)) / rate
            else:
                next_event_time = math.inf
        if (survival and pop > 0) or not survival:
            # This conditions on survival of the population for the
            # monitoring period
            break
    out[thread_id] = pop



import numpy as np
from scipy.linalg import expm
import birdepy.utility as ut


def probability_expm(z0, zt, t, param, b_rate, d_rate, z_trunc):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *matrix exponential* method.

    To use this function call :func:`birdepy.probability` with `method` set to
    'expm': ::

        birdepy.probability(z0, zt, t, param, method='expm', z_trunc=())

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------
    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default
        ``z_trunc=[z_min, z_max]`` where ``z_min=max(0, min(z0, zt) - 100)``
        and ``z_max=max(z0, zt) + 100``

    Examples
    --------
    Approximate transition probability for a Verhulst model using a matrix exponential: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='expm')[0][0]

    Outputs: ::

        0.0027414224836612463

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    # Extract the min and max population sizes considered
    z_min, z_max = z_trunc

    # If more than one time is requested it is easiest to divert into a different code block
    if t.size == 1:
        q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)
        p_mat = expm(np.multiply(q_mat, t))
        output = p_mat[np.ix_(np.array(z0 - z_min, dtype=np.int32),
                              np.array(zt - z_min, dtype=np.int32))]
    else:
        # Initialize an output array to be filled in as we loop over times
        output = np.zeros((t.size, z0.size, zt.size))
        # Determine the transition rate matrix
        q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)
        for idx in range(t.size):
            # The transition probabilities follow from taking a matrix exponential
            p_mat = expm(np.multiply(q_mat, t[idx]))
            # Fill in the output according to requested initial and final states (although many more
            # probabilities are actually computed)
            output[idx, :, :] = p_mat[np.ix_(np.array(z0 - z_min, dtype=np.int32),
                                             np.array(zt - z_min, dtype=np.int32))]
    return output



import numpy as np
from scipy.stats import norm


def probability_oua(z0, zt, t, param, b_rate, d_rate, h_fun, zf_bld):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *Ornstein-Uhlenbeck approximation* method.

    To use this function call :func:`birdepy.probability` with `method` set to
    'oua'::

        birdepy.probability(z0, zt, t, param, method='oua')

    This function does not have any arguments which are not already described
    by the documentation of :func:`birdepy.probability`

    Examples
    --------
    Approximate transition probability for a Verhulst model using Ornstein--Uhlenbeck approximation: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='oua')[0][0]

    Outputs: ::

        0.0018882966813798246

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    For a method related to this one see [3].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Ross, J.V., Taimre, T. and Pollett, P.K. On parameter estimation
     in population models. Theoretical Population Biology, 70(4):498-510, 2006.

    """

    zfs = np.array(zf_bld(param))

    zfs = zfs[~np.isnan(zfs)]

    h_vals = []
    for point in zfs:
        h_vals.append(h_fun(point, param))

    idx_min_h_val = np.argmin(h_vals)

    zf = zfs[idx_min_h_val]
    h = h_vals[idx_min_h_val]

    # If more than one time is requested it is easiest to divert into a different code block
    if t.size == 1:
        output = np.zeros((z0.size, zt.size))
    else:
        output = np.zeros((t.size, z0.size, zt.size))

    # A distribution over all possible final states is determined for each requested initial state
    for idx1, _z0 in enumerate(z0):
        # Determine the mean of the approximation distribution
        m = zf + np.multiply(np.exp(np.multiply(h, t)), (_z0 - zf))

        # Determine the variance of the approximation distribution
        s2 = np.multiply(
            np.divide(np.add(b_rate(zf, param), d_rate(zf, param)),
                      2 * h),
            np.exp(np.multiply(2 * h, t)) - 1)

        # Given the mean and covariance, use a normal distribution pdf to determine probabilities
        # for requested terminal states
        for idx2 in range(zt.size):
            if t.size == 1:
                output[idx1, idx2] = norm.pdf(zt[idx2], loc=m[0],
                                              scale=np.sqrt(max(s2[0],
                                                                1e-30)))
            else:
                for idx3 in range(t.size):
                    output[idx3, idx1, idx2] = \
                        norm.pdf(zt[idx2], loc=m[idx3],
                                 scale=np.sqrt(max(s2[idx3], 1e-30)))

    return output



import numpy as np


def sq_bld(data, b_rate, d_rate):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using the *Galton-Watson approximation least
    lse estimation* method.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'lse' and `squares` set to 'gwa': ::

        bd.estimate(t_data, p_data, p0, p_bounds, framework='lse', squares='gwa')

    This function does not have any arguments which are not already described
    by the documentation of ``birdepy.estimate`` (see :ref:`here <Estimation>`)
    or use ``help(birdepy.estimate)``.

    Examples
    --------


    See also
    --------
    birdepy.estimate
    birdepy.forecast

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    def error_fun(param):
        err = 0
        for i in data:
            z0 = i[0]
            zt = i[1]
            expected_pop = z0 * np.exp(
                (b_rate(z0, param) / z0 - d_rate(z0, param) / z0) * i[2])
            err += np.square(expected_pop - zt)
        return err

    return lambda p_prop: error_fun(p_prop)



import numpy as np
import birdepy.utility_probability as ut
import mpmath as mp
import warnings


def w_fun(lam, mu, _z0, _zt, t):
    """
    Computes an important part of the saddlepoint approximation for a linear
    linear-and-death process. Corresponds to $\tilde s = (2A)^{-1}(-B+\sqrt{B^2-4AC})$
    in Lemma 1 of reference [1].

    References
    ----------
    .. [1] Davison, A. C., Hautphenne, S., & Kraus, A. (2021). Parameter
    estimation for discretely observed linear birth‐and‐death processes.
    Biometrics, 77(1), 186-196.
    """
    if lam == mu:
        a = lam * t - (lam * t) ** 2
        b = 2 * (lam * t) ** 2 + _z0 / _zt - 1
        c = -lam * t - (lam * t) ** 2
    else:
        m = np.exp((lam - mu) * t)
        a = lam * (m - 1) * (lam - mu * m)
        b1 = 2 * lam * mu * (1 + m ** 2 - m - (_z0 / _zt) * m)
        b2 = m * (lam ** 2 + mu ** 2) * ((_z0 / _zt) - 1)
        b = b1 + b2
        c = mu * (m - 1) * (mu - lam * m)
    if a == 0:
        w = -c / b
    elif (b ** 2 - 4 * a * c) < 0:
        raise Exception
    elif b == 0:
        raise Exception
    else:
        w = (-b + np.sqrt(b ** 2 - 4 * a * c)) / (2 * a)

    return w


def w_fun_high_precision(lam, mu, _z0, _zt, t):
    """
    Computes an important part of the saddlepoint approximation for a linear
    linear-and-death process in high precision. Corresponds to
    $\tilde s = (2A)^{-1}(-B+\sqrt{B^2-4AC})$ in Lemma 1 of reference [1].

    References
    ----------
    .. [1] Davison, A. C., Hautphenne, S., & Kraus, A. (2021). Parameter
    estimation for discretely observed linear birth‐and‐death processes.
    Biometrics, 77(1), 186-196.
    """
    if lam == mu:
        a = mp.fsub(mp.fmul(lam, t),
                    mp.power(mp.fmul(lam, t), 2))
        b = mp.fsum((mp.fprod((2.0,
                               mp.power(lam, 2),
                               mp.power(t, 2))),
                     mp.fdiv(_z0, _zt),
                     -1.0))
        c = -mp.fadd(mp.fmul(lam, t),
                     mp.power(mp.fmul(lam, t), 2))
    else:
        m = mp.exp(mp.fmul(mp.fsub(lam, mu), t))
        a = mp.fprod((lam,
                      mp.expm1(mp.fmul(mp.fsub(lam, mu), t)),
                      mp.fsub(lam, mp.fmul(mu, m))))
        b1 = mp.fprod((2.0,
                       lam,
                       mu,
                       mp.fsub(1.0,
                               mp.fmul(m,
                                       mp.fadd(
                                           mp.expm1(mp.fmul(mp.fsub(lam, mu),
                                                            t)),
                                           mp.fdiv(_z0, _zt)
                                       )
                                       )
                               )
                       )
                      )
        b2 = mp.fprod((m,
                       mp.fadd(mp.power(lam, 2), mp.power(mu, 2)),
                       mp.fsub(mp.fdiv(_z0, _zt), 1)
                       )
                      )
        b = mp.fadd(b1, b2)
        c = mp.fprod((mu,
                      mp.expm1(mp.fmul(mp.fsub(lam, mu), t)),
                      mp.fsub(mu, mp.fmul(lam, m))))
    if a == 0 and b > 0:
        w = -mp.fdiv(c, b)
    elif mp.fsub(mp.power(b, 2), 4 * mp.fmul(a, c)) < 0:
        w = 1e-100
        warnings.warn("A value of b**2 - 4ac less than 0 has been "
                      "encountered and w was replaced by a default value. "
                      "The results may be unreliable.",
                      category=RuntimeWarning)
    elif b == 0:
        w = 1e-100
        warnings.warn("A value of b equal to 0 has been encountered"
                      "and w was replaced by a default value. The results"
                      "may be unreliable. ",
                      category=RuntimeWarning)
    else:
        w = mp.fdiv(mp.fadd(-b,
                            mp.sqrt(mp.fsub(mp.power(b, 2),
                                            mp.fprod((4.0, a, c))))),
                    mp.fmul(2.0, a))

    return w


def probability_gwasa(z0, zt, t, param, b_rate, d_rate, anchor):
    """Transition probabilities for continuous-time birth-and-death processes
    using the *Galton-Watson saddlepoint approximation* method.

    To use this function call :func:`birdepy.probability` with `method` set to
    'gwasa'::

        birdepy.probability(z0, zt, t, param, method='gwasa', anchor='midpoint')

    The parameters associated with this method (listed below) can be
    accessed using kwargs in :func:`birdepy.probability()`. See documentation
    of :func:`birdepy.probability` for the main arguments.

    Parameters
    ----------
    anchor : string, optional
        Determines which state z is used to determine the linear approximation.
        Should be one of: 'initial' (z0 is used), 'midpoint' (default, 0.5*(z0+zt) is used),
        'terminal' (zt is used), 'max' (max(z0, zt) is used), or 'min' (min(z0, zt) is used).

    Examples
    --------
    Approximate transition probability for a Verhulst model using the saddlepoint approximation
    to a linear approximation: ::

        import birdepy as bd
        bd.probability(19, 27, 1.0, [0.5, 0.3, 0.02, 0.01], model='Verhulst', method='gwasa')[0][0]

    Outputs: ::

        0.002271944691896704

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    # This function computes a transition probability for a specific initial state, terminal state
    # and time combination. Later we will use it in a loop.
    def prob(_z0, _zt, _t):
        # Set the rates of the linear approximation according the argument of 'anchor'
        if anchor == 'midpoint':
            midpoint = 0.5*(_z0+_zt)
            lam = b_rate(midpoint, param) / midpoint
            mu = d_rate(midpoint, param) / midpoint
        elif anchor == 'initial':
            lam = b_rate(_z0, param) / _z0
            mu = d_rate(_z0, param) / _z0
        elif anchor == 'terminal':
            lam = b_rate(_zt, param) / _zt
            mu = d_rate(_zt, param) / _zt
        elif anchor == 'max':
            max_z = max(_z0, _zt)
            lam = b_rate(max_z, param) / max_z
            mu = d_rate(max_z, param) / max_z
        elif anchor == 'min':
            min_z = min(_z0, _zt)
            lam = b_rate(min_z, param) / min_z
            mu = d_rate(min_z, param) / min_z
        else:
            raise TypeError("Argument 'anchor' has an unknown value. Should be one of 'midpoint'"
                            " 'initial', 'terminal', 'max' or 'min'.")
        # Linear birth-and-death processes never evolve away from 0
        if _z0 == 0 and _zt > 0:
            raise TypeError("Methods 'gwa' and 'gwasa' are not suitable for "
                            "datasets that include transitions away "
                            "from the origin (i.e., z_{i-1}=0 and "
                            "z_i>0). ")
        # Linear birth-and-death processes always stay at 0
        if _z0 == 0 and _zt == 0:
            pr = 1
        elif _zt == 0:  # Transitioning to 0 is a special case
            pr = mp.fmul(_z0, np.log(max(1e-100, ut.beta1(lam, mu, _t))))
        elif _zt == 1:  # Transitioning to 1 is a special case
            pr = ut.p_lin(_z0, 1, b_rate(_z0, param) / _z0,
                          d_rate(_z0, param) / _z0, _t)
        else:
            # We attempt to compute required quantities with machine precision. If an error is
            # encountered we divert to high precision computations (which take longer).
            # The computations performed here are described in detail in Section 2.3 of
            # reference [1]
            with np.errstate(divide='raise', over='raise', under='raise',
                             invalid='raise'):
                try:
                    w = w_fun(lam, mu, _z0, _zt, _t)
                except:
                    w = w_fun_high_precision(lam, mu, _z0, _zt, _t)
            if w == 0:
                w = 1e-100
                warnings.warn(
                    "A value of w equal to 0 has been encountered and "
                    " w was replaced by a default value. The results "
                    " may be unreliable.",
                    category=RuntimeWarning)            
            with np.errstate(divide='raise', over='raise', under='raise',
                             invalid='raise'):
                try:
                    if lam == mu:
                        p1 = lam * _t * (1 - w) + w
                        p2 = 1 - lam * _t * (w - 1)
                        p3 = lam * _t * w * (
                                -lam * _t * w ** 2 + lam * _t + w ** 2 + 1)
                        p4 = (lam * _t * (w - 1) - 1) ** 2 * (
                                -lam * _t * w + lam * _t + w) ** 2
                    else:
                        m = np.exp((lam - mu) * _t)
                        p1 = mu - lam * w + mu * (w - 1) * m
                        p2 = mu - lam * w + lam * (w - 1) * m
                        p3 = -(m - 1) * m * w * (lam - mu) ** 2 * (
                                -lam ** 2 * w ** 2 +
                                lam * m * mu * (w ** 2 - 1) +
                                mu ** 2)
                        p4 = (lam * (m * (w - 1) - w) + mu) ** 2 * \
                             (lam * w + mu * (-m * w + m - 1)) ** 2
                    pr = ((p1 / p2) ** _z0 * (p3 / p4) ** (-0.5)) / \
                         (np.sqrt(2 * np.pi * _z0) * w ** _zt)
                except:
                    mp.dps = 1000
                    w = mp.mpmathify(w)
                    if lam == mu:
                        p1 = mp.fadd(mp.fprod((lam,
                                               _t,
                                               mp.fsub(1.0, w))),
                                     w)
                        p2 = mp.fsub(1.0,
                                     mp.fprod((lam,
                                               _t,
                                               mp.fsub(w, 1))
                                              )
                                     )
                        p3 = mp.fprod((lam,
                                       _t,
                                       w,
                                       mp.fsum((mp.fprod((-lam,
                                                          _t,
                                                          mp.power(w, 2))),
                                                mp.fmul(lam, _t),
                                                mp.power(w, 2),
                                                1.0))
                                       ))
                        p4 = mp.fmul(
                            mp.power(mp.fsub(mp.fprod((lam,
                                                       _t,
                                                       mp.fsub(w, 1))),
                                             1.0), 2),
                            mp.power(mp.fsum((mp.fprod((-lam,
                                                        _t,
                                                        w)),
                                              mp.fmul(lam, _t),
                                              w)), 2)
                        )
                    else:
                        m = mp.exp(mp.fmul(mp.fsub(lam, mu), _t))
                        p1 = mp.fsum((mu,
                                      mp.fmul(-lam, w),
                                      mp.fprod((mu,
                                                mp.fsub(w, 1),
                                                m))
                                      ))
                        p2 = mp.fsum((mu,
                                      mp.fmul(-lam, w),
                                      mp.fprod((lam,
                                                mp.fsub(w, 1),
                                                m))))
                        p3 = -mp.fprod(
                            (mp.expm1(mp.fmul(mp.fsub(lam, mu), _t)),
                             m,
                             w,
                             mp.power(mp.fsub(lam, mu), 2),
                             mp.fsum((-mp.power(mp.fmul(lam, w), 2),
                                      mp.fprod((lam,
                                                m,
                                                mu,
                                                mp.fsub(
                                                    mp.power(w, 2),
                                                    1.0)
                                                )),
                                      mp.power(mu, 2)
                                      ))))
                        p4 = mp.fmul(
                            mp.power(
                                mp.fadd(mp.fmul(lam,
                                                mp.fsub(
                                                    mp.fmul(m,
                                                            mp.fsub(w, 1.0)),
                                                    w)),
                                        mu),
                                2),
                            mp.power(mp.fadd(mp.fmul(lam, w),
                                             mp.fmul(mu,
                                                     mp.fsum((-mp.fmul(m, w),
                                                              m,
                                                              -1.0)))),
                                     2))
                    pr = mp.fprod((mp.fdiv(1.0,
                                           mp.fmul(mp.sqrt(mp.fprod(
                                               (2.0, mp.pi, _z0))),
                                               mp.power(w, _zt))),
                                   mp.power(mp.fdiv(p1, p2), _z0),
                                   mp.power(mp.fdiv(p3, p4), -0.5)))
        # Return a float (not a high precision mpmath object)
        if type(pr) != float:
            pr = float(mp.re(pr))

        if not 0 <= pr <= 1.0:
            warnings.warn("Probability not in [0, 1] computed, "
                          "some output has been replaced by a "
                          "default value. "
                          " Results may be unreliable.",
                          category=RuntimeWarning)
            if pr < 0:
                pr = 0.0
            else:
                pr = 1.0

        return pr

    # If more than one time is requested it is easiest to divert into a different code block
    if t.size == 1:
        # Initialize an output array to be filled in as we loop over initial and terminal states
        output = np.zeros((z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            for idx2, _zt in enumerate(zt):
                output[idx1, idx2] = prob(_z0, _zt, t[0])
    else:
        # Initialize an output array to be filled in as we loop over times, initial and terminal
        # states
        output = np.zeros((t.size, z0.size, zt.size))
        for idx1, _z0 in enumerate(z0):
            for idx2, _zt in enumerate(zt):
                for idx3, _t in enumerate(t):
                    output[idx3, idx1, idx2] = prob(_z0, _zt, _t)
    return output



import numpy as np
import birdepy.utility as ut
from birdepy.interface_aug import f_fun_bld as f_fun_bld


def discrete_est_em_none(sorted_data, p0, likelihood, technique, known_p,
                         idx_known_p, model, b_rate, d_rate, z_trunc,
                         p_bounds, con, max_it, i_tol, j_tol, h_tol, display,
                         opt_method, options):
    """
    Executes an expectation-maximization algorithm (not accelerated), to estimate
    parameters for a population-size-dependent birth-and-death process.

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.

    [1] Jamshidian, M., & Jennrich, R. I. (1997). Acceleration of the EM algorithm
     by using quasi‐Newton methods. Journal of the Royal Statistical Society:
     Series B (Statistical Methodology), 59(3), 569-587.
    """
    #Initialize a list to return parameters in each iteration and store the initial parameters in it
    iterations = [p0.tolist()]
    it = 1
    p_est = p0
    difference = i_tol + 1
    while difference > i_tol and it < max_it:
        # Build surrogate log-likelihood function (defined in Section 2.1 of ref [1])
        f_fun = f_fun_bld(sorted_data, p_est, b_rate, d_rate, likelihood,
                          technique, idx_known_p, known_p, model, z_trunc,
                          j_tol, h_tol, options)[0]

        # Negate the surrogate log-likelihood function so that scipy.optimize.minimize()
        # Can be used to find a maximum
        def error_fun(p_prop):
            return -f_fun(p_prop)

        # Find the estimate for this iteration
        opt = ut.minimize_(error_fun, p_est, p_bounds, con, opt_method,
                           options)
        p_est_next = opt.x

        # Compute the difference between this iteration estimate and
        # previous iteration estimate
        difference = np.sum(np.abs(p_est_next - p_est))
        # Prepare for next iteration
        p_est = p_est_next
        # Store the estimate
        iterations.append(p_est.tolist())
        # Print a progress update if requested
        if display:
            print('Iteration ', it, ' estimate is: ', p_est)
        it += 1

    return list(p_est), iterations



import numpy as np
from birdepy import simulate
import birdepy.utility as ut
import matplotlib.pyplot as plt
from matplotlib import cm
import warnings
from scipy.integrate import solve_ivp


def parameter_sampler(param, cov, p_bounds, known_p, idx_known_p, con, rng):
    """
    Samples parameter values according to a truncated multivariate normal with
    expected value given by argument 'param', covariance given by argument 'cov'
    and truncated to support defined by 'p_bounds' and 'con'.
    """
    while True:
        # First obtain proposal satisfying bounds:
        while True:
            param_prop = rng.multivariate_normal(param, cov)
            cond = True
            for bd_idx, bd in enumerate(p_bounds):
                if param_prop[bd_idx] > bd[1] or \
                        param_prop[bd_idx] < bd[0]:
                    cond = False
            if not cond:
                continue
            break
        # Then check proposal also satisfies constraints:
        if type(con) == dict:
            # If there is one constraint specified
            if con['fun'](param_prop) < 0:
                continue
        else:
            # If there is strictly more than one or 0 constraints specified
            cond = True
            for c in con:
                if c['fun'](param_prop) < 0:
                    cond = False
            if not cond:
                continue
        break
    param_prop = ut.p_bld(np.array(param_prop), idx_known_p, known_p)
    return param_prop


def mean_curve(param, b_rate, d_rate, method, z0, solver_methods,
               shift_times, n, rng):
    """
    Generates an approximation to the expected value of a population-size-dependent
    birth-and-death process using a deterministic (fluid) approximation or
    simulation.
    """
    if method == 'fm':
        if callable(z0):
            z0_ = z0()
        else:
            z0_ = z0
        for solver_method in solver_methods:
            fluid_path = solve_ivp(
                lambda t, z: b_rate(z, param) - d_rate(z, param),
                [0, shift_times[-1]],
                [z0_],
                t_eval=shift_times,
                method=solver_method)
            if fluid_path.success:
                forecasted_mean = fluid_path.y[0]
                break
    elif method in ['exact', 'ea', 'ma', 'gwa']:
        forecasted_mean = np.zeros(len(shift_times))
        for idx in range(n):
            forecasted_mean += simulate.discrete(param, 'custom', z0,
                                                 shift_times,
                                                 b_rate=b_rate,
                                                 d_rate=d_rate, k=1,
                                                 method=method,
                                                 seed=rng)
        forecasted_mean = np.divide(forecasted_mean, n)
    return forecasted_mean


def forecast(model, z0, times, param, cov=None, interval='confidence', method=None,
             percentiles=(0, 2.5, 10, 25, 50, 75, 90, 97.5, 100),
             labels=('$95\%$', '$80\%$', '$50\%$'),
             p_bounds=None, con=(), known_p=(), idx_known_p=(),
             k=10 ** 3, n=10 ** 3, seed=None, colormap=cm.Purples,
             xlabel='t', ylabel='default', xticks='default',
             rotation=45, display=False, export=False, **options):
    """Simulation based forecasting for continuous-time birth-and-death processes.
    Produces a plot of the likely range of mean population sizes subject to parameter uncertainty
    (confidence intervals) or the likely range of population sizes subject to parameter
    uncertainty and model stochasticity (prediction intervals).

    Parameters
    ----------
    model : string, optional
        Model specifying birth and death rates of process (see :ref:`here
        <Birth-and-death Processes>`). Should be one of:

            - 'Verhulst' (default)
            - 'Ricker'
            - 'Hassell'
            - 'MS-S'
            - 'Moran'
            - 'pure-birth'
            - 'pure-death'
            - 'Poisson'
            - 'linear'
            - 'linear-migration'
            - 'M/M/1'
            - 'M/M/inf'
            - 'loss-system'
            - 'custom'

         If set to 'custom', then kwargs `b_rate` and `d_rate` must also be
         specified. See :ref:`here <Custom Models>` for more information.

    z0: int or callable
        The population for each sample path at the time of the first element
        of the argument of `times`.
        If it is a callable it should be a function that has no arguments and
        returns an int:

         ``z0() -> int``

    times : array_like
        Times to provide a forecast for.

    param : array_like
        The parameters governing the evolution of the birth-and-death
        process to be forecast.
        Array of real elements of size (m,), where ‘m’ is the number of
        parameters.
        When `cov` is provided this is taken to be a mean value.

    cov : array_like, optional
        The parameters are assumed to follow a truncated normal distribution
        with this covariance. If this is specified, then p_bounds should also be
        specified to avoid unwanted parameters.

    interval : string, optional
        Type of forecast. Should be one of 'confidence' (default) or
        'prediction'. Confidence intervals show the likely range of mean future
        population values, reflecting parameter uncertainty. Prediction interals
        show the likely range of future population values, incorporating
        parameter uncertainty and model stochasticity.

    method : string, optional
        Method used to generate samples. For confidence intervals samples are
        trajectories of future expected values. For prediction intervals
        samples are trajectories of future population values. Should be one of:

            - 'fm' (default for confidence intervals)
            - 'exact'
            - 'ea'
            - 'ma'
            - 'gwa' (default for prediction intervals)

    percentiles : list, optional
        List of percentiles to split the data into.

    labels : list, optional
        List of strings containing labels for each percentile split.

    p_bounds : list
        Bounds on parameters. Should be specified as a sequence of
        ``(min, max)`` pairs for each unknown parameter. See :ref:`here <Parameter Constraints>`.

    con : {Constraint, dict} or List of {Constraint, dict}, optional
        Constraints definition for parameters.
        See :ref:`here <Parameter Constraints>` for more information.

    known_p : array_like, optional
        List of known parameter values. For built in models these must be in
        their canonical order as given (:ref:`here <Birth-and-death Processes>`). If this
        argument is given, then argument `idx_known_p` must also be specified.
        See :ref:`here <Known Parameters>` for more information.

    idx_known_p : array_like, optional
        List of indices of known parameters (as given in argument 'known_p').
        For built in models indices must correspond to canonical order as given
        :ref:`here <Birth-and-death Processes>`. If this argument is given, then argument
        `known_p` must also be specified. See :ref:`here <Known Parameters>`
        for more information.

    k : int, optional
        Number of samples used to generate forecast. For confidence intervals
        each sample corresponds to an estimate of the mean for a sampled
        parameter value. For prediction intervals each sample corresponds to
        a trajectory of population size for a sampled parameter value.

    n : int, optional
        Number of samples used to estimate each sample of a mean for confidence
        interval samples. Only applicable when method is 'exact', 'ea', 'ma'
        or 'gwa'.

    seed : int, Generator, optional
        If *seed* is not specified the random numbers are generated according
        to *np.random.default_rng()*. If *seed* is an *int*, random numbers are
        generated according to *np.random.default_rng(seed)*. If seed is a
        *Generator*, then that object is used. See
        :ref:`here <Reproducibility>` for more information.

    colormap : matplotlib.colors.LinearSegmentedColormap, optional
        Colors used for plot.

    xlabel : str, optional
        Label for x axis of plot.

    ylabel : str, optional
        Label for y axis of plot.

    xticks : array_like, optional
        Locations of x ticks.

    rotation : int, optional
        Rotation of x tick labels.

    display : bool, optional
        If True, then progress updates are provided.

    export : str, optional
        File name for export of the figure to a tex file.

    Examples
    --------
    First simulate some sample paths using : :`birdepy.simulate.discrete()`: ::

        import birdepy as bd
        t_data = list(range(101))
        p_data = bd.simulate.discrete([0.75, 0.25, 0.02, 1], 'Ricker', 10, t_data,
                                      survival=True, seed=2021)

    Then, using the simulated data, estimate the parameters: ::

        est = bd.estimate(t_data, p_data, [0.5, 0.5, 0.05], [[0,1], [0,1], [0, 0.1]],
                          model='Ricker', idx_known_p=[3], known_p=[1])

    Then, use the estimated parameters and covariances to generate a forecast: ::

        future_t = list(range(101,151,1))
        bd.forecast('Ricker', p_data[-1], future_t, est.p, cov=est.cov,
                    p_bounds=[[0,1], [0,1], [0, 0.1]], idx_known_p=[3], known_p=[1],
                    interval='prediction')


    Notes
    -----
    This function creates a plot but does not return anything.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    If `model` is 'Poisson', then the true dnm is immediately returned (i.e.,
    the total number of events during the observation periods divided by the
    total observation time).

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    :func:`birdepy.gpu_functions.probability()`  :func:`birdepy.gpu_functions.discrete()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    # Convert times into a numpy array
    times = np.array(times)

    # Prepare for plotting
    if type(xticks) == str:
        xticks = times
    if type(xticks) is list:
        xticks = times
    else:
        xticks = np.array(xticks)
    if ylabel == 'default':
        if interval == 'confidence':
            ylabel = '$\mathbb{E} Z(t)$'
        else:
            ylabel = '$Z(t)'

    # Determine method for computing sample paths
    if interval == 'confidence' and method is None:
        method = 'fm'
    elif interval == 'prediction' and method is None:
        method = 'gwa'
    elif interval == 'prediction' and method == 'fm':
        TypeError("Argument of `method` equal 'fm' not possible when argument "
                  "of `interval` equals 'prediction'.")

    # Shift the times so that the earliest observation is time 0
    shift_times = times - times[0]

    # Covariance matrix must be an array, convert to array if it is a float or int
    if type(cov) == float or type(cov) == int:
        cov = [[cov]]

    # Define the model-dependent birth and death rate functions
    if model == 'custom':
        b_rate = options['b_rate']
        d_rate = options['d_rate']
    else:
        b_rate = ut.higher_birth(model)
        d_rate = ut.higher_death(model)

    # Set the random number generator
    if seed is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(seed)

    # These are the methods we try to solve differential equations with in scipy.integrate.solve_ivp
    solver_methods = ['RK45', 'Radau', 'RK23', 'BDF', 'DOP853']

    # Initialise a plot
    fig, (ax1) = plt.subplots(nrows=1, ncols=1, figsize=(8, 4))

    # A difference procedure needs to be followed depending on whether a
    # 'confidence' or 'prediction' forecast is requested
    if interval == 'confidence':
        # Confidence region is not possible if no covariance matrix is provided
        if cov is None:
            warnings.warn("Confidence intervals show the likely range of the mean "
                          "future population level given the uncertainty of the "
                          "parameter values, however since argument `cov` has value "
                          "None no uncertainty in parameter values has been specified.",
                          category=RuntimeWarning)
            forecast_ = mean_curve(param, b_rate, d_rate, method, z0,
                                   solver_methods, shift_times, n, rng)
            ax1.plot(times, forecast_, color='k')
        else:
            # Create an array to store samples in
            samples = np.zeros((k, times.shape[0]))
            for idx in range(k):
                # Sample new parameters
                param_prop = parameter_sampler(param, cov, p_bounds, known_p,
                                               idx_known_p, con, rng)
                # Generate a mean curve corresponding to the sampled parameters
                samples[idx, :] = mean_curve(param_prop, b_rate, d_rate,
                                             method, z0, solver_methods,
                                             shift_times, n, rng)
                # Print a progress indicator (if requested)
                if display:
                    print(f"Forecast is ", 100 * (idx + 1) / k, f"% complete.")
    elif interval == 'prediction':
        # Create an array to store samples in
        samples = np.zeros((k, times.shape[0]))
        # If no covariance matrix is provided the prediction interval only accounts for model
        # stochasticity
        if cov is None:
            for idx in range(k):
                # Generate sample path of process using estimated parameters
                samples[idx, :] = simulate.discrete(param, 'custom', z0,
                                                    shift_times,
                                                    b_rate=b_rate,
                                                    d_rate=d_rate, k=1,
                                                    method=method,
                                                    seed=rng)
                if display:
                    print(f"Forecast is ", 100 * (idx + 1) / k, f"% complete.")
        else:
            for idx in range(k):
                param_prop = parameter_sampler(param, cov, p_bounds, known_p,
                                               idx_known_p, con, rng)
                samples[idx, :] = simulate.discrete(param_prop, 'custom', z0,
                                                    shift_times,
                                                    b_rate=b_rate,
                                                    d_rate=d_rate, k=1,
                                                    method=method,
                                                    seed=rng)
                if display:
                    print(f"Forecast is ", 100 * (idx + 1) / k, f"% complete.")
    else:
        raise TypeError("Argument 'interval' has an unknown value.")

    # Determine number of partitions to split the samples into
    m = len(percentiles)
    # Determine percentile curves
    SDist = np.zeros((times.shape[0], m))
    for i in range(m):
        for t in range(times.shape[0]):
            SDist[t, i] = np.percentile(samples[:, t], percentiles[i])
    half = int(np.floor((m - 1) / 2))
    # Plot the results
    fig.canvas.draw()
    ax1.plot(times, SDist[:, half], color='k')
    for i in range(half-1):
        ax1.fill_between(times, SDist[:, i+1], SDist[:, -(i + 2)], color=colormap((i+1) / half),
                         label=labels[i])
    ax1.tick_params(labelsize=11.5)
    ax1.set_xlabel(xlabel, fontsize=14)
    ax1.set_xticks(xticks)
    ax1.legend(loc="upper left")
    labels = [f'{t}' for t in xticks]
    ax1.set_xticklabels(labels, rotation=rotation)
    ax1.set_ylabel(ylabel, fontsize=14)
    fig.tight_layout()
    # If requested a tex file can be created
    if isinstance(export, str):
        import tikzplotlib
        tikzplotlib.save(export + ".tex")
    return 0



import numpy as np
import birdepy.utility as ut
import birdepy.simulate as simulate
from scipy.spatial.distance import euclidean
from scipy.stats import multivariate_normal, norm
from scipy.optimize import minimize, minimize_scalar
import matplotlib.pyplot as plt


def estimate_prior_pdf(p_bounds, con, rng):
    """
    Uses Monte Carlo samples to provide an estimate of the density function
    for a uniform distribution with support defined by 'p_bounds' and 'con'.

    This function is used by :func:`discrete_est_abc()`.
    """
    p_bounds = np.array(p_bounds)
    successes = 0
    for idx in range(10 ** 3):
        param_prop = (p_bounds[:, 1] - p_bounds[:, 0]) * \
                     rng.uniform(size=p_bounds.shape[0]) + \
                     p_bounds[:, 0]
        success = True
        if type(con) == dict:
            # If there is one constraint specified
            if con['fun'](param_prop) < 0:
                success = False
        # If there is strictly more than one or exactly 0 constraints specified
        else:
            for c in con:
                if c['fun'](param_prop) < 0:
                    success = False
                    break
        successes += success
    return 1 / ((successes / 10 ** 3) * np.prod(p_bounds[:, 1] - p_bounds[:, 0]))


def param_sampler(previous_sample, k, weights, cov, p_bounds, con, rng):
    cond2 = True
    while cond2:
        cond2 = False
        # If there are no previous samples, sample a proposal satisfying the bounds
        # using a uniform distribution. Otherwise sample a proposal according to the
        # mechanism described in [3].
        if len(previous_sample) == 0:
            param_prop = (p_bounds[:, 1] - p_bounds[:, 0]) * \
                         rng.uniform(size=p_bounds.shape[0]) + p_bounds[:, 0]
        else:
            cond1 = True
            while cond1:
                cond1 = False
                index_mean = rng.choice(list(range(k)), p=np.divide(weights, np.sum(weights)))
                mean = previous_sample[index_mean, :]
                if p_bounds.shape[0] > 1:
                    param_prop = rng.multivariate_normal(mean, 2 * cov)
                else:
                    param_prop = rng.normal(mean, 2 * cov)
                for bnd_idx, bnd in enumerate(p_bounds):
                    if param_prop[bnd_idx] > bnd[1] or param_prop[bnd_idx] < bnd[0]:
                        cond1 = True
        # Then check proposal also satisfies constraints:
        if type(con) == dict:
            # If there is one constraint specified
            if con['fun'](param_prop) < 0:
                cond2 = True
        else:
            # If there is strictly more than one or 0 constraints specified
            for c in con:
                if c['fun'](param_prop) < 0:
                    cond2 = True
                    break  # Leave for loop on first failure
    return param_prop


def simulated_data_maker(data, param_prop, idx_known_p, known_p, b_rate,
                         d_rate, method, tau, rng):
    """
    Generates a simulated version of 'data' according to 'param_prop'.
    """
    param_full = ut.p_bld(param_prop, idx_known_p, known_p)
    data_pairs = np.empty((sum(data.values()), 2))
    lower_index = 0
    for i in data:
        # Here a transition is used where z0 = i[0], zt = i[1], t = i[2] and
        # the number of times the transition is observed is data[i]
        sample = simulate.discrete(param_full, model='custom',
                                   b_rate=b_rate, d_rate=d_rate, z0=i[0],
                                   times=[0, i[2]], k=data[i], method=method,
                                   tau=tau, seed=rng)
        if data[i] == 1:
            data_pairs[lower_index, :] = [i[1], sample[1]]
        else:
            data_pairs[lower_index:(lower_index + data[i]), 0] = i[1]
            data_pairs[lower_index:(lower_index + data[i]), 1] = sample[:, 1]
        lower_index += data[i]
    return data_pairs


def basic_abc(data, previous_sample, weights, cov, eps_abc, distance, k,
              method, b_rate, d_rate, idx_known_p, known_p, p_bounds, con, tau,
              rng, iteration, display):
    """
    Performs a single iteration of the basic ABC algorithm.
    """
    param_samples = np.empty((k, p_bounds.shape[0]))
    dist_samples = np.empty(k)
    for idx in range(k):
        while True:
            param_prop = param_sampler(previous_sample, k, weights, cov, p_bounds,
                                       con, rng)
            data_pairs = simulated_data_maker(data, param_prop, idx_known_p,
                                              known_p, b_rate, d_rate, method,
                                              tau, rng)
            dist = distance(data_pairs[:, 0], data_pairs[:, 1])
            if dist <= eps_abc:
                param_samples[idx, :] = param_prop
                dist_samples[idx] = dist
                break
        if display:
            print(f"Iteration {iteration} is {100 * (idx + 1) / k}% complete.")
    return param_samples, dist_samples


def discrete_est_abc(data, eps_abc, distance, stat, k, gam, max_its, max_q,
                     eps_change, method, b_rate, d_rate, idx_known_p, known_p,
                     p_bounds, con, tau, seed, display):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using approximate Bayesian computation.
    See :ref:`here <Approximate Bayesian Computation>` for more information.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'abc'::

        bd.estimate(t_data, p_data, p0, p_bounds, framework='abc', eps_abc='dynamic', k=100,
                    max_its=3, max_q=0.99, eps_change=5, gam=5, method='gwa', tau=None, seed=None,
                    distance=None, stat='mean', display=False)

    The parameters associated with this framework (listed below) can be
    accessed using kwargs in :func:`birdepy.estimate()`. See documentation of
    :func:`birdepy.estimate` for the main arguments.

    Parameters
    ----------
    eps_abc : list, str, optional
        Threshold distance betweeen simulated data and observed data for accepting
        parameter proposals. If set to 'dynamic' (default), then the procedure
        described in [3] is used. Otherwise `eps_abc` must be a list which
        specifies epsilon for each iteration.

    k : int, optional
        Number of successful parameter samples used to obtain estimate.

    max_its : int, optional
        Maximum number of iterations of algorithm.

    max_q : scalar, optional
        Tolerance threshold for stopping algorithm (see Equation 2.5 in [3]).
        Is only checked after at least two iterations have occurred.

    eps_change : scalar, optional
        An iteration is only performed if the percentage decrease in 'eps_abc'
        compared to the previous iteration is greater than this value.

    gam : int, optional
        If `eps_abc` is set to 'dynamic', then k*gam samples are initially
        sampled and the distance between the data and the k-th largest of these
        samples is used as the first value of epsilon

    method : string, optional
        Simulation algorithm used to generate samples (see
        :ref:`here<Simulation Algorithms>`). Should be one of:

            - 'exact'
            - 'ea'
            - 'ma'
            - 'gwa' (default)

    tau : scalar, optional
        Time between samples for the approximation methods 'ea', 'ma' and 'gwa'.
        Has default value ``min(x/10, 0.1)`` where 'x' is the smallest
        inter-observation time.

    seed : int, Generator, optional
        If `seed` is not specified the random numbers are generated according
        to :func:`np.random.default_rng()`. If `seed` is an 'int;, random numbers are
        generated according to :func:`np.random.default_rng(seed)`. If `seed` is a
        'Generator', then that object is used. See
        :ref:`here <Reproducibility>` for more information.

    distance : callable, optional
        Computes the distance between simulated data and observed data. Default
        value is :func:`scipy.spatial.distance.euclidean`.

    stat : string, optional
        Determines which statistic is used to summarize the posterior distribution.
        Should be one of: 'mean' or 'median'. 

    Examples
    --------
    Simulate a sample path and estimate the parameters using ABC: ::

        import birdepy as bd
        t_data = list(range(100))
        p_data = bd.simulate.discrete([0.75, 0.25, 0.02, 1], 'Ricker', 10, t_data,
                                      survival=True, seed=2021)

    Assume that the death rate and population size are known, then estimate the rate of spread: ::

        est = bd.estimate(t_data, p_data, [0.5], [[0,1]], framework='abc',
                          model='Ricker', idx_known_p=[1, 2, 3],
                          known_p=[0.25, 0.02, 1], display=True, seed=2021)
        print(f"abc estimate is {est.p}, with standard errors {est.se},
              computed in {est.compute_time} seconds.")

    Outputs: ::

        abc estimate is [0.72434934654868] , with standard errors [0.0520378066100896] computed in  122.48563146591187 seconds.

    Notes
    -----
    Estimation algorithms and models are also described in [1]. If you use this
    function for published work, then please cite this paper.

    For a text book treatment on the theory of birth-and-death processes
    see [2].

    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    .. [3] Simola, U., Cisewski-Kehe, J., Gutmann, M. U., & Corander, J. (2021).
     Adaptive approximate Bayesian computation tolerance selection. Bayesian analysis,
     16(2), 397-423.
    """
    if distance is None:
        distance = euclidean

    prior_pdf = estimate_prior_pdf(p_bounds, con, seed)

    known_p = np.array(known_p)
    p_bounds = np.array(p_bounds)

    list_of_samples = []
    param_sample = []
    weights = [1 / k] * k
    cov = []
    mean = []
    previous_mean = []
    previous_cov = []
    iteration = 0
    q = 0

    current_eps = np.inf
    while True:
        iteration += 1
        if iteration > max_its:
            break
        previous_eps = current_eps
        # First set the value of epsilon (the cut off for acceptance of parameter proposals)
        if eps_abc == 'dynamic':
            if iteration == 1:
                # param_sample = np.empty((gam * k, p_bounds.shape[0]))
                dist_samples = np.empty(gam * k)
                for idx in range(gam * k):
                    # The first argument of param_sample is [] to ensure a uniform dist is used
                    param_prop = param_sampler([], k, weights, cov, p_bounds,
                                               con, seed)
                    data_pairs = simulated_data_maker(data, param_prop, idx_known_p,
                                                      known_p, b_rate, d_rate, method,
                                                      tau, seed)
                    dist = distance(data_pairs[:, 0], data_pairs[:, 1])
                    # param_sample[idx, :] = param_prop
                    dist_samples[idx] = dist
                indices_of_smallest_distances = np.argpartition(dist_samples, k)
                dist_samples = dist_samples[indices_of_smallest_distances[:k]]
                # previous_sample = param_sample[indices_of_smallest_distances[:k], :]
                current_eps = np.amax(dist_samples)
            elif iteration == 2:
                if p_bounds.shape[0] > 1:
                    opt = minimize(lambda p: -multivariate_normal.pdf(p, mean, cov) / prior_pdf,
                                   x0=mean, method='SLSQP', bounds=p_bounds,
                                   constraints=con)
                else:
                    opt = minimize(
                        lambda p: -norm.pdf(p, loc=mean[0], scale=np.sqrt(cov)) / prior_pdf,
                        mean[0], method='SLSQP', bounds=p_bounds)
                q = -1 / opt.fun
                current_eps = np.quantile(dist_samples, q)
            else:
                if p_bounds.shape[0] > 1:
                    opt = minimize(
                        lambda p: -multivariate_normal.pdf(p, mean, cov) /
                                  max(1e-30,
                                      multivariate_normal.pdf(p, previous_mean,
                                                              previous_cov)),
                        x0=mean, method='SLSQP', bounds=p_bounds,
                        constraints=con)
                else:
                    opt = minimize(lambda p: -norm.pdf(p, loc=mean[0], scale=np.sqrt(cov)) /
                                             max(1e-30, norm.pdf(p, loc=previous_mean[0],
                                                                 scale=np.sqrt(
                                                                     previous_cov))),
                                   mean[0], method='SLSQP', bounds=p_bounds)
                q = -1 / opt.fun
                current_eps = np.quantile(dist_samples, q)
        else:
            current_eps = eps_abc[iteration - 1]
        if ((q > max_q and iteration >= 3) or
                (100 * (previous_eps / current_eps - 1) < eps_change)):
            break
        # Now run the basic ABC algorithm for this iteration
        previous_sample = param_sample
        param_sample, dist_samples = \
            basic_abc(data, previous_sample, weights, cov, current_eps, distance,
                      k, method, b_rate, d_rate, idx_known_p, known_p, p_bounds,
                      con, tau, seed, iteration, display)
        list_of_samples.append(param_sample)
        # create a temporary array (since the new values depend on the old ones)
        previous_weights = weights
        previous_cov = cov
        previous_mean = mean
        # Determine weights for next iteration
        if iteration == 1:
            weights = [1 / k] * k
            if p_bounds.shape[0] > 1:
                cov = np.cov(param_sample, rowvar=False, aweights=previous_weights)
            else:
                cov = np.var(np.multiply(param_sample.T, previous_weights))
        else:
            if p_bounds.shape[0] > 1:
                cov = np.cov(param_sample, rowvar=False, aweights=previous_weights)
                sd = np.sqrt(np.diag(cov))
                weights = np.empty(k)
                for i in range(k):
                    temp = 0
                    for j in range(k):
                        temp += previous_weights[j] * \
                                multivariate_normal.pdf(
                                    np.divide(param_sample[i, :] - previous_sample[j, :], sd),
                                    cov=np.eye(p_bounds.shape[0]))
                    weights[i] = prior_pdf / temp
            else:
                cov = np.var(np.multiply(param_sample.T, previous_weights))
                sd = np.sqrt(cov)
                weights = np.empty(k)
                for i in range(k):
                    temp = 0
                    for j in range(k):
                        temp += previous_weights[j] * \
                                norm.pdf((param_sample[i] -
                                          previous_sample[j]) / sd)
                    weights[i] = prior_pdf / temp
        mean = np.mean(param_sample, axis=0)
    if stat == 'mean':
        est = np.mean(param_sample, axis=0)
    else:
        est = np.median(param_sample, axis=0)
    if p_bounds.shape[0] > 1:
        cov = np.cov(param_sample, rowvar=False)
    else:
        cov = np.array([np.var(param_sample)])
    return est, cov, list_of_samples



import numpy as np
from scipy.linalg import expm
import birdepy.utility as ut


def sq_bld(data, b_rate, d_rate, z_trunc):
    """Parameter estimation for discretely observed continuous-time
    birth-and-death processes using the *matrix exponential least lse
    estimation* method.

    To use this function call :func:`birdepy.estimate` with `framework` set to
    'lse' and `squares` set to 'expm': ::

        bd.estimate(t_data, p_data, p0, p_bounds, framework='lse', squares='expm', z_trunc=())

    See documentation of ``birdepy.estimate`` (see :ref:`here <Estimation>`)
    or use ``help(birdepy.estimate)`` for the rest of the arguments.

    Parameters
    ----------
    z_trunc : array_like, optional
        Truncation thresholds, i.e., minimum and maximum states of process
        considered. Array of real elements of size (2,) by default
        ``z_trunc=[z_min, z_max]`` where
        ``z_min=max(0, z_obs_min - obs_range)`` and
        ``z_max=z_obs_max + obs_range`` with ``obs_range`` equal to the
        absolute difference between the highest and lowest observed
        populations.

    Examples
    --------


    See also
    --------
    :func:`birdepy.estimate()` :func:`birdepy.probability()` :func:`birdepy.forecast()`

    :func:`birdepy.simulate.discrete()` :func:`birdepy.simulate.continuous()`

    References
    ----------
    .. [1] Hautphenne, S., & Patch, B. (2021). Birth-and-death Processes in Python:
     The BirDePy Package. arXiv preprint arXiv:2110.05067.

    .. [2] Feller, W. (1968) An introduction to probability theory and its
     applications (Volume 1) 3rd ed. John Wiley & Sons.

    """
    # Extract the min and max population sizes considered
    z_min, z_max = z_trunc

    # Sort the data into a more efficient format
    sorted_data = ut.data_sort_2(data)

    def error_fun(param):
        # Determine the transition rate matrix
        q_mat = ut.q_mat_bld(z_min, z_max, param, b_rate, d_rate)
        err = 0
        for t in sorted_data:
            # The transition probabilities follow from taking a matrix exponential
            p_mat = expm(np.multiply(q_mat, t))
            for i in sorted_data[t]:
                expected_pop = np.dot(np.arange(z_min, z_max + 1, 1),
                                      p_mat[i[0] - z_min, :])
                err += sorted_data[t][i] * np.square(expected_pop - i[1])
        return err
    return lambda p_prop: error_fun(p_prop)




import mpmath as mp
import numpy as np
import birdepy.utility as ut
from birdepy.probability_ilt import laplace_p
from birdepy.interface_probability import probability
import birdepy.utility_em as ut_em


def aug_bld_ilt(sorted_data, param, b_rate, d_rate, likelihood, model, z_trunc,
                j_tol, h_tol, eps, laplace_method, num_terms, options):
    """
    Creates a function based on numerical Laplace transform inversion
    that returns the expected values of the number of up jumps from a
    state z, the number of down jumps from z, and the time spent in z
    when the process transitions from z0 to zt in elapsed time t.

    This function is called by :func:`bd.interface_aug.f_fun_bld()`.
    """
    def udh(z0, zt, t, z):
        """
        Computes the expected values defined by Equation (14) in
        reference [1] using numerical Laplace inversion.

        [1] Crawford, F. W., Minin, V. N., & Suchard, M. A. (2014).
        Estimation for general birth-death processes. Journal of
        the American Statistical Association, 109(506), 730-747.

        """
        pr = probability(z0, zt, t, param, model, likelihood, z_trunc=z_trunc,
                         options=options)[0][0]

        if b_rate(z, param) == 0 or pr == 0.0:
            u = 0.0
        else:
            pre_u = ut.laplace_invert(
                lambda s: laplace_p(s, z0, z, param, b_rate, d_rate, eps) *
                          laplace_p(s, z + 1, zt, param, b_rate, d_rate, eps),
                t,
                laplace_method=laplace_method,
                k=num_terms,
                f_bounds=[z_trunc[0], z_trunc[1]])
            u = float(mp.fdiv(mp.fmul(pre_u, b_rate(z, param)), pr))

        if d_rate(z, param) == 0 or pr == 0.0:
            d = 0.0
        else:
            pre_d = ut.laplace_invert(
                lambda s: laplace_p(s, z0, z, param, b_rate, d_rate, eps) *
                          laplace_p(s, z - 1, zt, param, b_rate, d_rate, eps),
                t,
                laplace_method=laplace_method,
                k=num_terms,
                f_bounds=[z_trunc[0], z_trunc[1]])
            d = float(mp.fdiv(mp.fmul(pre_d, d_rate(z, param)), pr))

        if pr == 0.0:
            h = 0.0
        else:
            pre_h = ut.laplace_invert(
                lambda s: laplace_p(s, z0, z, param, b_rate, d_rate, eps) *
                          laplace_p(s, z, zt, param, b_rate, d_rate, eps),
                t,
                laplace_method=laplace_method,
                k=num_terms,
                f_bounds=[0, t + 0.05])
            h = float(mp.fdiv(pre_h, pr))

        return np.array([u, d, h])

    aug_data = ut_em.help_bld_aug(udh, sorted_data, j_tol, h_tol, z_trunc)

    return aug_data
